"use strict";

var _httpErrors = _interopRequireDefault(require("http-errors"));

var _graphql = require("graphql");

var _expressGraphql = _interopRequireDefault(require("express-graphql"));

var _renderGraphiQL = require("./renderGraphiQL");

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

const {
  getGraphQLParams
} = _expressGraphql.default;
/**
 * Used to configure the graphqlHTTP middleware by providing a schema
 * and other configuration options.
 *
 * Options can be provided as an Object, a Promise for an Object, or a Function
 * that returns an Object or a Promise for an Object.
 */

/**
 * Middleware for express; takes an options object or function as input to
 * configure behavior, and returns an express middleware.
 */
module.exports = graphqlHTTP;

function graphqlHTTP(options) {
  if (options == null) {
    throw new Error('GraphQL middleware requires options.');
  }

  return async function middleware(ctx) {
    const req = ctx.req;
    const request = ctx.request;
    const response = ctx.response; // Higher scoped variables are referred to at various stages in the
    // asynchronous state machine below.

    let schema;
    let context;
    let rootValue;
    let fieldResolver;
    let typeResolver;
    let pretty;
    let graphiql;
    let formatErrorFn = _graphql.formatError;
    let validateFn = _graphql.validate;
    let executeFn = _graphql.execute;
    let parseFn = _graphql.parse;
    let extensionsFn;
    let showGraphiQL = false;
    let query;
    let documentAST;
    let variables;
    let operationName;
    let validationRules;
    let result;

    try {
      // Promises are used as a mechanism for capturing any thrown errors during
      // the asynchronous process below.
      // Resolve the Options to get OptionsData.
      const optionsData = await (typeof options === 'function' ? options(request, response, ctx) : options); // Assert that optionsData is in fact an Object.

      if (!optionsData || typeof optionsData !== 'object') {
        throw (0, _httpErrors.default)(500, 'GraphQL middleware option function must return an options object or a promise which will be resolved to an options object.');
      } // Assert that schema is required.


      if (!optionsData.schema) {
        throw (0, _httpErrors.default)(500, 'GraphQL middleware options must contain a schema.');
      } // Collect information from the options data object.


      schema = optionsData.schema;
      context = optionsData.context ?? ctx;
      rootValue = optionsData.rootValue;
      fieldResolver = optionsData.fieldResolver;
      typeResolver = optionsData.typeResolver;
      validationRules = optionsData.validationRules ?? [];
      pretty = optionsData.pretty;
      graphiql = optionsData.graphiql;
      extensionsFn = optionsData.extensions;

      if (optionsData.formatError) {
        // eslint-disable-next-line no-console
        console.warn('`formatError` is deprecated and replaced by `customFormatErrorFn`. It will be removed in version 1.0.0.');
      }

      validateFn = optionsData.customValidateFn ?? validateFn;
      executeFn = optionsData.customExecuteFn ?? executeFn;
      parseFn = optionsData.customParseFn ?? parseFn;
      formatErrorFn = optionsData.customFormatErrorFn ?? optionsData.formatError ?? formatErrorFn; // GraphQL HTTP only supports GET and POST methods.

      if (request.method !== 'GET' && request.method !== 'POST') {
        throw (0, _httpErrors.default)(405, 'GraphQL only supports GET and POST requests.', {
          headers: {
            Allow: 'GET, POST'
          }
        });
      } // Use request.body when req.body is undefined.


      req.body = req.body || request.body; // Parse the Request to get GraphQL request parameters.

      const params = await getGraphQLParams(req); // Get GraphQL params from the request and POST body data.

      query = params.query;
      variables = params.variables;
      operationName = params.operationName;
      showGraphiQL = canDisplayGraphiQL(request, params) && graphiql;
      result = await new Promise(resolve => {
        // If there is no query, but GraphiQL will be displayed, do not produce
        // a result, otherwise return a 400: Bad Request.
        if (query == null) {
          if (showGraphiQL) {
            resolve(null);
            return;
          }

          throw (0, _httpErrors.default)(400, 'Must provide query string.');
        } // Validate Schema


        const schemaValidationErrors = (0, _graphql.validateSchema)(schema);

        if (schemaValidationErrors.length > 0) {
          // Return 500: Internal Server Error if invalid schema.
          throw (0, _httpErrors.default)(500, 'GraphQL schema validation error.', {
            graphqlErrors: schemaValidationErrors
          });
        } // GraphQL source.


        const source = new _graphql.Source(query, 'GraphQL request'); // Parse source to AST, reporting any syntax error.

        try {
          documentAST = parseFn(source);
        } catch (syntaxError) {
          // Return 400: Bad Request if any syntax errors errors exist.
          throw (0, _httpErrors.default)(400, 'GraphQL syntax error.', {
            graphqlErrors: [syntaxError]
          });
        } // Validate AST, reporting any errors.


        const validationErrors = validateFn(schema, documentAST, [..._graphql.specifiedRules, ...validationRules]);

        if (validationErrors.length > 0) {
          // Return 400: Bad Request if any validation errors exist.
          throw (0, _httpErrors.default)(400, 'GraphQL validation error.', {
            graphqlErrors: validationErrors
          });
        } // Only query operations are allowed on GET requests.


        if (request.method === 'GET') {
          // Determine if this GET request will perform a non-query.
          const operationAST = (0, _graphql.getOperationAST)(documentAST, operationName);

          if (operationAST && operationAST.operation !== 'query') {
            // If GraphiQL can be shown, do not perform this query, but
            // provide it to GraphiQL so that the requester may perform it
            // themselves if desired.
            if (showGraphiQL) {
              resolve(null);
              return;
            } // Otherwise, report a 405: Method Not Allowed error.


            throw (0, _httpErrors.default)(405, `Can only perform a ${operationAST.operation} operation from a POST request.`, {
              headers: {
                Allow: 'POST'
              }
            });
          }
        } // Perform the execution, reporting any errors creating the context.


        try {
          resolve(executeFn({
            schema,
            document: documentAST,
            rootValue,
            contextValue: context,
            variableValues: variables,
            operationName,
            fieldResolver,
            typeResolver
          }));
          response.status = 200;
        } catch (contextError) {
          // Return 400: Bad Request if any execution context errors exist.
          throw (0, _httpErrors.default)(400, 'GraphQL execution context error.', {
            graphqlErrors: [contextError]
          });
        }
      }); // Collect and apply any metadata extensions if a function was provided.
      // https://graphql.github.io/graphql-spec/#sec-Response-Format

      if (result && extensionsFn) {
        // eslint-disable-next-line require-atomic-updates
        result = await Promise.resolve(extensionsFn({
          document: documentAST,
          variables,
          operationName,
          result,
          context
        })).then(extensions => {
          if (extensions && typeof extensions === 'object') {
            result.extensions = extensions;
          }

          return result;
        });
      }
    } catch (error) {
      // If an error was caught, report the httpError status, or 500.
      // eslint-disable-next-line require-atomic-updates
      response.status = error.status ?? 500;

      if (error.headers != null) {
        for (const [key, value] of Object.entries(error.headers)) {
          response.set(key, value);
        }
      } // eslint-disable-next-line require-atomic-updates


      result = {
        errors: error.graphqlErrors ?? [error]
      };
    } // If no data was included in the result, that indicates a runtime query
    // error, indicate as such with a generic status code.
    // Note: Information about the error itself will still be contained in
    // the resulting JSON payload.
    // https://graphql.github.io/graphql-spec/#sec-Data


    if (response.status === 200 && result && !result.data) {
      response.status = 500;
    } // Format any encountered errors.


    if (result && result.errors) {
      result.errors = result.errors.map(err => formatErrorFn(err, context));
    } // If allowed to show GraphiQL, present it instead of JSON.


    if (showGraphiQL) {
      const payload = (0, _renderGraphiQL.renderGraphiQL)({
        query,
        variables,
        operationName,
        result,
        options: typeof showGraphiQL !== 'boolean' ? showGraphiQL : {}
      });
      response.type = 'text/html';
      response.body = payload;
    } else {
      // Otherwise, present JSON directly.
      const payload = pretty ? JSON.stringify(result, null, 2) : result;
      response.type = 'application/json';
      response.body = payload;
    }
  };
}
/**
 * Helper function to determine if GraphiQL can be displayed.
 */


function canDisplayGraphiQL(request, params) {
  // If `raw` false, GraphiQL mode is not enabled.
  // Allowed to show GraphiQL if not requested as raw and this request prefers HTML over JSON.
  return !params.raw && request.accepts(['json', 'html']) === 'html';
}